import os
import sys

import pytest

from stag.config import Config
from stag.site import Site


sys.path.append(os.path.join(os.path.dirname(__file__), "helpers"))


@pytest.fixture
def config():
    return Config()


@pytest.fixture
def site(config):
    site = Site(config=config)
    return site



from stag.ecs import Path, Page, Metadata


_NO_TYPE = "_NO_TYPE"


def make_meta(*types, content=""):
    for t in types:
        if t is _NO_TYPE:
            yield Metadata()
        else:
            yield Metadata(type=t)


def make_pages(site, *metadatas):
    # not set to ease assertions
    for i, md in enumerate(metadatas):
        yield site.make_page(f"/{i}", metadata=md)


def test_page_filter_any(site):
    pages = list(
        make_pages(site, *make_meta(None, "page", "page", "post", _NO_TYPE), None)
    )
    filtered = set(site.filter_pages())
    assert filtered == set(site.pages[0:5])


def test_page_filter_none(site):
    pages = list(
        make_pages(site, *make_meta(None, "page", "page", "post", _NO_TYPE), None)
    )
    filtered = set(site.filter_pages(None))
    assert filtered == {site.pages[0]}


def test_page_filter_type(site):
    pages = list(
        make_pages(site, *make_meta(None, "page", "page", "post", _NO_TYPE), None)
    )
    filtered = set(site.filter_pages("page"))
    assert filtered == {site.pages[1], site.pages[2]}


def test_page_filter_unexisting_type(site):
    pages = list(
        make_pages(site, *make_meta(None, "page", "page", "post", _NO_TYPE), None)
    )
    filtered = set(site.filter_pages("someethingelse"))
    assert filtered == set()



import os
import shutil
from multiprocessing import Process

import pytest

from stag.stag import main_
from stag.utils import chdir
from stag.config import read_config
from stag.signals import signals

SITES_DIR = os.path.join(os.path.dirname(__file__), "sites")
DISABLED = []


def copy_site(name, dst):
    source = os.path.join(SITES_DIR, name)
    shutil.copytree(source, dst)


def get_sites():
    _, directories, _ = next(os.walk(SITES_DIR))
    return directories


def get_files(root):
    for rd, _, fs in os.walk(root):
        for f in fs:
            yield os.path.join(rd, f)


@pytest.mark.parametrize("sitename", get_sites())
def test_site_generation(sitename, tmp_path):
    if sitename in DISABLED:
        pytest.skip(f"Disabled test for site generation: {sitename}")

    site_dir = tmp_path / sitename
    copy_site(sitename, site_dir)
    with chdir(site_dir):
        print(site_dir)
        output = "_output"

        # Use multiprocessing to sandbox each function call from each other.
        # Things which have global state which might need clearing include, but
        # are not limited to:
        #   - cached read_config()
        #   - global signals
        #   - loaded plugins (this is the tough one)
        p = Process(target=main_, args=(["build", "-o", output],))
        p.start()
        p.join()

        with open("expected_files.txt") as ef:
            expected_files = sorted(line.strip() for line in ef)

        generated_files = sorted(f[len(f"{output}/") :] for f in get_files(output))

        assert generated_files == expected_files



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
from datetime import datetime

import tomli

from stag.ecs import Page, Content, Metadata
from stag.plugins._helpers import read_file


def is_xml(page):
    return page.source and page.source.ext == "xml"


def read(page):
    if not is_xml(page):
        return

    md, cn, _ = read_file(page.source.path)
    md.setdefault("title", page.source.filebase.capitalize())
    md.setdefault("type", page.source.filebase)
    md.setdefault("date", datetime.now())
    page.metadata = Metadata(md)
    page.input = Content("xml", cn)
    page.output = Content("xml", cn)


def register_plugin(site):
    site.signals.page_added.connect(read)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
from fnmatch import fnmatch
from urllib.parse import urljoin

from typing import Optional as _Optional
from collections.abc import Mapping as _Mapping

import attr

from stag.ecs import Page
from stag.signals import signal, Signals
from stag.reading import Readers


# sentinel used for parameters for which None is a valid value
_ANY = "_ANY"


def _urlize(url):
    url = url.strip("/")
    return f"/{url}"


def _absurl(base, path):
    return urljoin(base, path)


@attr.s(auto_attribs=True)
class Site:
    config: dict = attr.ib(repr=False)
    _pages: _Mapping[str, Page] = attr.ib(factory=dict, repr=False)
    _signals: Signals = attr.ib(init=False, factory=Signals, repr=False)
    _readers: Readers = attr.ib(init=False, factory=Readers, repr=False)

    @property
    def signals(self):
        return self._signals

    @property
    def readers(self):
        return self._readers

    @property
    def pages(self):
        return list(self._pages.values())

    @property
    def taxonomies(self):
        for page in self.pages:
            if page.taxonomy:
                yield page

    @property
    def ordinary_pages(self):
        for page in self._pages.values():
            if page.source and page.input and page.output and page.metadata:
                yield page

    def subpages_of(self, val, recursive=False):
        def _cmp(pardir, exp):
            return pardir == exp

        def _cmp_r(pardir, exp):
            cp = os.path.commonpath([pardir, exp])
            return cp == exp

        cmp_fn = _cmp_r if recursive else _cmp

        base = _urlize(val)
        for page in self.ordinary_pages:
            pardir = os.path.dirname(page.relurl)
            if cmp_fn(pardir, base):
                yield page

    def make_page(self, path, **kw):
        path = _urlize(path)
        if path in self._pages:
            raise ValueError(f"URL {path} already exists")

        url = _absurl(self.config.url, path)
        page = Page(url, **kw)
        self._pages[path] = page
        self.signals.page_added.emit(page)
        return page

    def get_or_make_page(self, path, **kw):
        path = _urlize(path)
        if path in self._pages:
            return self._pages[path]

        url = _absurl(self.config.url, path)
        page = Page(url, **kw)
        self._pages[path] = page
        self.signals.page_added.emit(page)
        return page

    def filter_pages(self, ptype=_ANY):
        for page in self.pages:
            if page.metadata is not None and ptype is _ANY:
                yield page
            elif page.metadata is not None and page.metadata.get("type", _ANY) == ptype:
                yield page

    def find(self, path):
        return self._pages.get(_urlize(path))

    def resources(self, page, basename_glob=None):
        for other in self.pages:
            if (
                other != page
                and other.source
                and page.source
                and other.source.reldirname == page.source.reldirname
            ) and (
                basename_glob is None or fnmatch(other.source.basename, basename_glob)
            ):
                yield other


@attr.s(auto_attribs=True)
class SiteTemplateProxy:
    _site: Site

    def __getattr__(self, name):
        forbidden = {"signals", "readers", "make_page", "get_or_make_page"}
        if name in forbidden:
            raise AttributeError(f"Access to {name} is forbidden from inside template")
        return getattr(self._site, name)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import weakref
from functools import wraps as _wraps
from typing import List as _List

import attr


@attr.s(auto_attribs=True)
class signal:
    name: str
    _observers: _List[weakref.ref] = attr.ib(init=False, factory=list, repr=False)

    def connect(self, slot, weak=True):
        if weak:
            ref = weakref.ref(slot)
        else:
            ref = slot

        if ref not in self._observers:
            self._observers.append(ref)

    def disconnect(self, slot, weak=True):
        try:
            if weak:
                self._observers.remove(weakref.ref(slot))
            else:
                self._observers.remove(slot)
        except ValueError:
            pass

    def emit(self, *args, **kwargs):
        missing = []

        for i, ref in enumerate(self._observers):
            obj = ref() if isinstance(ref, weakref.ref) else ref
            if obj:
                obj(*args, **kwargs)
            else:
                missing.append(i)

        for i in reversed(missing):
            del self._observers[i]

    def clear(self):
        self._observers.clear()


def condition(cond):
    def decor(fn):
        @_wraps(fn)
        def wrapper(*a, **kw):
            if cond(*a, **kw):
                return fn(*a, **kw)

        return wrapper

    return decor


class Signals:
    def __init__(self):
        # some predefined signals
        self.register_signal("plugins_loaded")
        self.register_signal("site_finished")
        self.register_signal("readers_init")
        self.register_signal("readers_finished")
        self.register_signal("processors_init")
        self.register_signal("processors_finished")
        self.register_signal("rendering_init")
        self.register_signal("rendering_finished")
        self.register_signal("page_added")

    def register_signal(self, name):
        if hasattr(self, name):
            return getattr(self, name)

        s = signal(name)
        setattr(self, name, s)
        return s

    def clear(self):
        for sig in self.__dict__:
            getattr(self, sig).clear()


signals = Signals()



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

from datetime import datetime
from collections import defaultdict

from slugify import slugify

from stag.signals import signals
from stag.ecs import Page, Taxonomy, Term, Metadata, Content
from stag.utils import chdir, SortDate

_TAXONOMIES = []
_RAW_TAXO = defaultdict(lambda: defaultdict(set))


def copy_output_type(pages, default="html"):
    for page in pages:
        if page.output and page.output.type:
            return page.output.type
    return default


def update_metadata(page, **kw):
    if not page.metadata:
        page.metadata = Metadata()
    kw.update(page.metadata.data)
    page.metadata.data = kw


def init_taxonomies(site):
    if not site.config.taxonomies:
        return

    global _TAXONOMIES
    _TAXONOMIES = [
        Taxonomy(name=ttab.key, singular=ttab.singular, plural=ttab.plural)
        for ttab in site.config.taxonomies
    ]


def add_pages(site):
    for page in site.pages:
        if not page.metadata:
            continue

        for taxonomy in _TAXONOMIES:
            terms = page.metadata.get(taxonomy.name)

            if not terms:
                continue
            if not isinstance(terms, list):
                terms = [terms]

            for term in terms:
                _RAW_TAXO[taxonomy.name][term].add(page)


def finalize(site):
    # Pages can be created by users for taxonomies and terms. In this situation
    # stag will override some members (page.list), update others with unset
    # values (metadata) and leave the rest untouched (e.g. content).
    #
    # This is to give users possibility to hardcode custom metadata for specific
    # taxonomies.
    now = datetime.now()
    for taxonomy in _TAXONOMIES:
        baseslug = slugify(taxonomy.name)
        taxo_page = site.get_or_make_page(baseslug)
        taxo_page.taxonomy = taxonomy
        update_metadata(taxo_page, title=taxonomy.plural, date=now)

        for term, pages in _RAW_TAXO[taxonomy.name].items():
            termslug = slugify(term)
            term_page = site.get_or_make_page(f"{baseslug}/{termslug}")
            term_page.term = Term(termslug, sorted(pages, key=SortDate()))
            term_page.output = Content(type=copy_output_type(pages))
            update_metadata(term_page, title=term, date=now, taxonomy=taxonomy.name)

            taxo_page.taxonomy.terms.append(term_page)

        taxo_page.taxonomy.terms.sort()
        taxo_page.output = Content(type=copy_output_type(taxo_page.taxonomy.terms))


def register_plugin(site):
    global _TAXONOMIES
    global _RAW_TAXO
    _TAXONOMIES = []
    _RAW_TAXO = defaultdict(lambda: defaultdict(set))

    site.signals.readers_finished.connect(add_pages)
    site.signals.processors_init.connect(finalize)
    init_taxonomies(site)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import asyncio
from asyncio.subprocess import PIPE
import os
import logging

import attr

from stag.ecs import Page, Content, Metadata
from stag.plugins._helpers import read_file

log = logging.getLogger(__name__)


ADOC_EXTENSIONS = {"adoc", "asc", "asciidoc"}


def get_process_limit():
    count = os.cpu_count()
    if count is None:
        return 2
    return count


@attr.s(auto_attribs=True)
class AsciidocConfig:
    process_limit: int = get_process_limit()


def is_adoc(page):
    return page.source and page.source.ext in ADOC_EXTENSIONS


def is_opened_md(page):
    return page.input and page.input.type == "asciidoc"


def deduce_url(path):
    if path.filebase == "index":
        return path.reldirname
    return os.path.join(path.reldirname, path.filebase)


async def asciidoctor(content):
    proc = await asyncio.create_subprocess_exec(
        "asciidoctor",
        "--no-header-footer",
        "--out-file",
        "-",
        "-",
        stdout=PIPE,
        stdin=PIPE,
    )
    stdout, _ = await proc.communicate(content.encode())
    html = stdout.decode()
    return html


async def generate_page(page):
    if not "title" in page.metadata:
        log.error(f"No title in {page.source.relpath}")
        return

    html = await asciidoctor(page.input.content)
    page.output = Content("html", html)


def chunks(iterable, size):
    if size == 0:
        yield iterable
        return

    for i in range(0, len(iterable), size):
        yield iterable[i : i + size]


async def generate_async(site):
    myconfig = site.config.plugins.asciidoc

    adocs = []
    for page in site.pages:
        if not is_opened_md(page):
            continue
        adocs.append(generate_page(page))

    for chunk in chunks(adocs, myconfig.process_limit):
        await asyncio.gather(*chunk)


def read(page):
    if not is_adoc(page):
        return

    metadata, content, _ = read_file(page.source.path)
    page.metadata = Metadata(metadata)
    page.input = Content("asciidoc", content)


def generate(site):
    asyncio.run(generate_async(site))


def register_plugin(site):
    site.config.update_plugin_table("asciidoc", AsciidocConfig())
    site.signals.page_added.connect(read)
    site.signals.processors_init.connect(generate)
    site.readers.register_reader(deduce_url, lambda p: p.ext in ADOC_EXTENSIONS)



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
import sys
import glob
import queue
import multiprocessing as mp
import traceback
from contextlib import contextmanager
from http.server import HTTPServer, SimpleHTTPRequestHandler
import logging

from stag.utils import chdir

log = logging.getLogger(__name__)


class Process(mp.Process):
    def __init__(self, *a, **kw):
        self._queue = mp.Queue()
        super().__init__(*a, **kw)

    def run(self):
        try:
            super().run()
        except Exception as e:
            tb = traceback.format_exc()
            self._queue.put((e, tb))

    @property
    def exception(self):
        try:
            e = self._queue.get_nowait()
            return e
        except queue.Empty:
            return None


# This function is intended to be run in a separate process
def run_http_server(directory, port):
    with chdir(directory):
        log.info(f"Running simple HTTP server on http://localhost:{port}.")
        log.info(f"Serving files from '{directory}'.")
        server_address = ("", port)
        httpd = HTTPServer(server_address, SimpleHTTPRequestHandler)
        httpd.serve_forever()


def get_timestamps(watch_patterns):
    stamps = {}
    for pat in watch_patterns:
        for path in glob.iglob(pat, recursive=True):
            mtime = os.stat(path).st_mtime
            stamps[path] = mtime
    return stamps


@contextmanager
def run_server(serve_directory, port):
    server = Process(target=run_http_server, args=(serve_directory, port))
    try:
        server.start()
        yield server
    finally:
        log.info("Terminating HTTP server.")
        server.terminate()
        server.join()
        server.close()


def serve_until_changed(serve_directory, port, watch_patterns):
    poll_delay = 1

    stamps = get_timestamps(watch_patterns)
    fl = len(stamps)
    log.info(f"Watching {fl} files.")

    with run_server(serve_directory, port) as server:
        while True:
            newstamps = get_timestamps(watch_patterns)
            if stamps != newstamps:
                return

            server.join(poll_delay)

            e = server.exception
            if e:
                raise e[0]

            if server.exitcode is not None:
                return



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
import itertools

from typing import List

import attr
import markdown

from stag.ecs import Page, Content, Metadata
from stag.plugins._helpers import read_file


def is_md(page):
    return page.source and page.source.ext == "md"


def is_opened_md(page):
    return page.input and page.input.type == "md"


@attr.s(auto_attribs=True)
class MarkdownConfig:
    extensions: List[str] = attr.Factory(lambda: ["footnotes", "fenced_code", "smarty"])


def deduce_url(path):
    if path.filebase == "index":
        return path.reldirname
    return os.path.join(path.reldirname, path.filebase)


def read(page):
    if not is_md(page):
        return

    metadata, content, _ = read_file(page.source.path)
    page.metadata = Metadata(metadata)
    page.input = Content("md", content)


def generate(site):
    myconfig = site.config.plugins.markdown
    conv = markdown.Markdown(extensions=myconfig.extensions)

    for page in site.pages:
        if not is_opened_md(page):
            continue

        assert "title" in page.metadata, f"No title in {page.source.relpath}"
        html = conv.reset().convert(page.input.content)
        page.output = Content("html", html)


def register_plugin(site):
    site.config.update_plugin_table("markdown", MarkdownConfig())
    site.signals.page_added.connect(read)
    site.signals.processors_init.connect(generate)
    site.readers.register_reader(deduce_url, lambda p: p.ext == "md")



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
import logging

import attr
from jinja2 import Template, Environment, FileSystemLoader, FunctionLoader, ChoiceLoader

from .template_functions import update_env
from stag.site import SiteTemplateProxy

try:
    from stag._version import version
except ImportError:
    version = "unknown version"


log = logging.getLogger(__name__)


_DEFAULT_TEMPLATE = "{{ content }}"
_DEFAULT_HTML_TEMPLATE = f"""<!DOCTYPE html>
<html lang="{{{{ site.config.get('lang', 'en') }}}}">
  <head>
    <meta name="generator" content="stag ({version})" />
    <meta charset="UTF-8">
    <title>{{{{ page.metadata.title }}}} - {{{{ site.config.title }}}}</title>
  </head>
  <body>
  {{{{ content }}}}
  </body>
</html>
"""


def load_default_template(name):
    log.error(f"Template not found: {name}. Using built-in basic template.")
    _, ext = os.path.splitext(name)
    template = _DEFAULT_HTML_TEMPLATE if ext == ".html" else _DEFAULT_TEMPLATE
    return template


def get_env(theme, config):
    env = Environment(
        loader=ChoiceLoader(
            [
                FileSystemLoader(theme),
                FunctionLoader(load_default_template),
            ]
        ),
        extensions=["jinja2.ext.debug", "jinja2.ext.do"],
        trim_blocks=True,
        lstrip_blocks=True,
    )
    update_env(env, config)
    return env


def is_term(page):
    return bool(page.term)


def is_taxonomy(page):
    return bool(page.taxonomy)


def get_default_type(page, templates):
    if is_taxonomy(page):
        return templates["taxonomy"]
    if is_term(page):
        return templates["list"]
    return templates["page"]


def file_in_url(url):
    bn = os.path.basename(url)
    _, ext = os.path.splitext(bn)
    return bool(ext)


def render_page(page, site, env):
    config = site.config
    myconfig = config.template

    type_ = page.metadata.get("type")
    if not type_:
        type_ = get_default_type(page, myconfig.templates)

    template = env.get_template(f"{type_}.{page.output.type}")
    url = page.relurl.strip("/")

    # TODO: test this behavior (mgoral, 2021-09-21)
    if file_in_url(url):
        outpath = os.path.join(config.output, url)
    else:
        index = f"index.{page.output.type}"
        outpath = os.path.join(config.output, url, index)

    if os.path.exists(outpath):
        log.error(f"page already exists, skipping: {outpath}")
        return

    outdir = os.path.dirname(outpath)
    if outdir:
        os.makedirs(outdir, exist_ok=True)
    with open(outpath, "w") as fd:
        sp = SiteTemplateProxy(site)
        fd.write(template.render(content=page.output.content, site=sp, page=page))



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

from __future__ import annotations

import os
from collections import UserDict, defaultdict
from datetime import date as _date_t
from datetime import datetime as _datetime_t
from urllib.parse import urlparse
from functools import cache

from typing import Optional as _Optional
from typing import Any as _Any

import attr
from dateutil.parser import parse as _parse_dt

from stag.signals import signal


@attr.s(auto_attribs=True, frozen=True)
class Path:
    path: str
    root_dir: str

    @property
    def relpath(self):
        return self.path[len(self.root_dir) :].strip("/")

    @property
    def ext(self):
        return os.path.splitext(self.path)[1][len(os.extsep) :]

    @property
    def filebase(self):
        return os.path.splitext(self.basename)[0]

    @property
    def basename(self):
        return os.path.basename(self.path)

    @property
    def reldirname(self):
        return os.path.dirname(self.relpath)


@attr.s(auto_attribs=True)
class Content:
    type: str
    content: _Any = attr.ib(None, repr=False)


class Metadata(UserDict):
    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._normalize_data()

    def _normalize_data(self):
        title = self.data.get("title", "")

        date = self.data.get("date")
        if date:
            if isinstance(date, (int, str, float)):
                self.data["date"] = _parse_dt(date)
            elif type(date) == _date_t:
                self.data["date"] = _datetime_t.fromordinal(date.toordinal())
            self.data["timestamp"] = self.data["date"].timestamp()

        lastmod = self.data.get("lastmod")
        if lastmod:
            if isinstance(lastmod, (int, str, float)):
                self.data["lastmod"] = _parse_dt(lastmod)
            elif type(lastmod) == _date_t:
                self.data["lastmod"] = _datetime_t.fromordinal(lastmod.toordinal())
            self.data["lastmod_timestamp"] = self.data["lastmod"].timestamp()

        self.__dict__.update(self.data)


@attr.s(auto_attribs=True)
class Taxonomy:
    name: str
    singular: str
    plural: str
    terms: _List[Page] = attr.ib(factory=list, repr=False)


@attr.s(auto_attribs=True)
class Term:
    name: str
    pages: _List[Page] = attr.ib(factory=list)


class _Notified:
    """Descriptor of fields which should automatically notify some of their
    changes."""

    def __init__(self, name):
        self.name = name
        self.real = "_" + name

    def __get__(self, obj, objtype):
        return getattr(obj, self.real)

    def __set__(self, obj, val):
        curr = getattr(obj, self.real)
        setattr(obj, self.real, val)

        # emit signals
        if val is None and curr is not None:
            signame = f"{self.name}_removed"
            getattr(obj, signame).emit(obj, curr)
        elif val != curr:
            signame = f"{self.name}_created"
            getattr(obj, signame).emit(obj, val)


def component(type_, **kw):
    kw.setdefault("metadata", {})["component"] = True
    kw.setdefault("repr", False)
    return attr.ib(None, **kw)


def component_descriptors(cls):
    for a in cls.__attrs_attrs__:
        if a.metadata and a.metadata.get("component"):
            dname = a.name.lstrip("_")
            setattr(cls, dname, _Notified(dname))

    return cls


@component_descriptors
@attr.s(auto_attribs=True, cmp=False, hash=False)
class Page:
    _url: str = attr.ib(converter=urlparse)

    _metadata: _Optional[Metadata] = component(Metadata)
    _source: _Optional[Path] = component(Path)
    _input: _Optional[Content] = component(Content)
    _output: _Optional[Content] = component(Content)
    _taxonomy: _Optional[Taxonomy] = component(Taxonomy)
    _term: _Optional[Term] = component(Term)

    def __attrs_post_init__(self):
        for a in self.__attrs_attrs__:
            dname = a.name.lstrip("_")
            if a.metadata and a.metadata.get("component"):
                crsig = f"{dname}_created"
                setattr(self, crsig, signal(crsig))
                rmsig = f"{dname}_removed"
                setattr(self, rmsig, signal(rmsig))

    @property
    @cache
    def url(self):
        return self._url.geturl()

    @property
    @cache
    def relurl(self):
        return self._url.path

    @property
    def md(self):
        return self._metadata

    def __hash__(self):
        return hash(self._url)

    def __eq__(self, other):
        return self._url == other._url

    def __lt__(self, other):
        return self._url < other._url



import pytest

from stag.ecs import Page, Metadata, Path
from stag.writers.template_functions import sorted_pages


@pytest.fixture
def files():
    return [
        Page("/static.txt"),
        Page("/bar", metadata=Metadata(date="2021-08-09")),
        Page("/empty", metadata=Metadata()),
        Page("/foo", metadata=Metadata(date="2021-08-10")),
        Page("/baz", metadata=Metadata(date="2020-11-11")),
    ]


def test_sort_default_key(files):
    srt = sorted_pages(files)
    assert srt == [files[1], files[4], files[2], files[3], files[0]]


def test_sort_default_key_reverse(files):
    srt = sorted_pages(files, reverse=True)
    assert srt == [files[0], files[3], files[2], files[4], files[1]]


def test_sort_date(files):
    srt = sorted_pages(files, key="date")
    assert srt[0] == files[4]
    assert srt[1] == files[1]
    assert srt[2] == files[3]
    # last 2 elements don't have date, so they are considered equal and their
    # order is undefined


def test_sort_date_reverse(files):
    srt = sorted_pages(files, key="date", reverse=True)
    # first 2 elements don't have date, so they are considered equal and their
    # order is undefined
    assert srt[2] == files[3]
    assert srt[3] == files[1]
    assert srt[4] == files[4]



# SPDX-License-Identifier: GPL-3.0-or-later
# Copyright (C) 2021 Michał Góral.

import os
import copy
from functools import cache

from typing import Optional, Any

import attr
import tomli

from stag.utils import attr_access


class ConfigError(Exception):
    pass


@attr.s(auto_attribs=True)
class TemplateTable:
    name: str = os.path.join("themes", "default")
    templates: dict = attr.ib(factory=dict, converter=attr_access)

    def __attrs_post_init__(self):
        self.templates.setdefault("page", "page")
        self.templates.setdefault("list", "list")
        self.templates.setdefault("taxonomy", "taxonomy")


@attr.s(auto_attribs=True)
class TaxonomyTable:
    key: str
    singular: Optional[str] = None
    plural: Optional[str] = None

    def __attrs_post_init__(self):
        if self.singular is None:
            self.singular = self.key
        if self.plural is None:
            self.plural = self.key


def init_table(cls):
    def _init(val):
        if isinstance(val, dict):
            return cls(**val)
        if isinstance(val, (tuple, list)):
            return type(val)(cls(**elem) for elem in val)
        if isinstance(val, cls):
            return val
        return cls(val)

    return _init


@attr.s(auto_attribs=True)
class Config:
    title: str = "MySite"
    url: str = "https://example.com"
    language: str = "en"
    timezone: str = "+0000"
    plugins_path: str = "plugins"
    plugins_disabled: list[str] = attr.ib(factory=list)
    content: str = "content"
    output: str = "_output"
    taxonomies: list[TaxonomyTable] = attr.ib(
        factory=list, converter=init_table(TaxonomyTable)
    )
    template: TemplateTable = attr.ib(
        factory=TemplateTable, converter=init_table(TemplateTable)
    )
    user: Any = attr.ib(factory=dict, converter=attr_access)
    plugins: Any = attr.ib(factory=dict, converter=attr_access)

    def get(self, name, default=None):
        return getattr(self, name, default)

    def update_plugin_table(self, name, table):
        tc = copy.deepcopy(table)
        if name in self.plugins:
            tc.__dict__.update(self.plugins[name])
            self.plugins[name] = tc.__dict__
        else:
            self.plugins[name] = tc.__dict__


@cache
def read_config(path):
    try:
        with open(path) as f:
            config_dct = tomli.loads(f.read())
    except FileNotFoundError:
        config_dct = {}

    return Config(**config_dct)
